# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import mock
import requests
import time
import unittest

import six

from monasca_notification import notification as m_notification
from monasca_notification.plugins import pagerduty_notifier

if six.PY2:
    import Queue as queue
else:
    import queue


def alarm(metrics):
    return {"tenantId": "0",
            "alarmId": "0",
            "alarmName": "test Alarm",
            "oldState": "OK",
            "newState": "ALARM",
            "severity": "LOW",
            "link": "some-link",
            "lifecycleState": "OPEN",
            "stateChangeReason": "I am alarming!",
            "timestamp": time.time(),
            "metrics": metrics}


class requestsResponse(object):
    def __init__(self, status):
        self.status_code = status


class TestWebhook(unittest.TestCase):
    def setUp(self):
        self.trap = queue.Queue()
        self.pagerduty_config = {'timeout': 50, 'key': 'foobar'}

    def tearDown(self):
        self.assertTrue(self.trap.empty())

    def _http_post_200(self, url, data, headers, **kwargs):
        self.trap.put(url)
        self.trap.put(data)
        self.trap.put(headers)
        r = requestsResponse(200)
        return r

    def _http_post_201(self, url, data, headers, **kwargs):
        self.trap.put(url)
        self.trap.put(data)
        self.trap.put(headers)
        r = requestsResponse(201)
        return r

    def _http_post_202(self, url, data, headers, **kwargs):
        r = requestsResponse(202)
        return r

    def _http_post_204(self, url, data, headers, **kwargs):
        self.trap.put(url)
        self.trap.put(data)
        self.trap.put(headers)
        r = requestsResponse(204)
        return r

    def _http_post_400(self, url, data, headers, **kwargs):
        r = requestsResponse(400)
        return r

    def _http_post_403(self, url, data, headers, **kwargs):
        r = requestsResponse(403)
        return r

    def _http_post_404(self, url, data, headers, **kwargs):
        r = requestsResponse(404)
        return r

    def _http_post_500(self, url, data, headers, **kwargs):
        r = requestsResponse(500)
        return r

    def _http_post_504(self, url, data, headers, **kwargs):
        r = requestsResponse(504)
        return r

    def _http_post_exception(self, url, data, headers, **kwargs):
        self.trap.put("timeout %s" % kwargs["timeout"])
        raise requests.exceptions.Timeout

    def valid_pagerduty_message(self, url, data, headers):
        self.assertEqual(
            url, 'https://events.pagerduty.com/generic/2010-04-15/create_event.json')

        headers = dict(headers)
        self.assertEqual(headers['content-type'], 'application/json')

        data = dict(json.loads(data))
        self.assertEqual(data['service_key'], 'ABCDEF')
        self.assertEqual(data['event_type'], 'trigger')
        self.assertEqual(data['description'], 'I am alarming!')
        self.assertEqual(data['client'], 'Monasca')
        self.assertEqual(data['client_url'], '')

        details = dict(data['details'])
        self.assertEqual(details['alarm_id'], '0')
        self.assertEqual(details['alarm_name'], 'test Alarm')
        self.assertEqual(details['current'], 'ALARM')
        self.assertEqual(details['message'], 'I am alarming!')

    def pagerduty_http_error(self, log_msg, http_response):
        self.assertRegexpMatches(log_msg, "Error with pagerduty request.")
        self.assertRegexpMatches(log_msg, "key=<ABCDEF>")
        self.assertRegexpMatches(log_msg, "response=%s" % http_response)

    @mock.patch('monasca_notification.plugins.pagerduty_notifier.requests')
    def notify(self, http_func, mock_requests):
        mock_log = mock.MagicMock()
        mock_log.warn = self.trap.put
        mock_log.error = self.trap.put
        mock_log.exception = self.trap.put

        mock_requests.post = http_func

        pagerduty = pagerduty_notifier.PagerdutyNotifier(mock_log)

        pagerduty.config(self.pagerduty_config)

        metric = []
        metric_data = {'dimensions': {'hostname': 'foo1', 'service': 'bar1'}}
        metric.append(metric_data)

        alarm_dict = alarm(metric)

        notification = m_notification.Notification(0,
                                                   'pagerduty',
                                                   'pagerduty notification',
                                                   'ABCDEF',
                                                   0,
                                                   0,
                                                   alarm_dict)

        self.trap.put(pagerduty.send_notification(notification))

    def test_pagerduty_200(self):
        """pagerduty 200
        """

        self.notify(self._http_post_200)

        url = self.trap.get(timeout=1)
        data = self.trap.get(timeout=1)
        headers = self.trap.get(timeout=1)

        self.valid_pagerduty_message(url, data, headers)

        return_value = self.trap.get()
        self.assertTrue(return_value)

    def test_pagerduty_201(self):
        """pagerduty 201
        """

        self.notify(self._http_post_201)

        url = self.trap.get(timeout=1)
        data = self.trap.get(timeout=1)
        headers = self.trap.get(timeout=1)

        self.valid_pagerduty_message(url, data, headers)

        return_value = self.trap.get()
        self.assertTrue(return_value)

    def test_pagerduty_204(self):
        """pagerduty 204
        """

        self.notify(self._http_post_204)

        url = self.trap.get(timeout=1)
        data = self.trap.get(timeout=1)
        headers = self.trap.get(timeout=1)

        self.valid_pagerduty_message(url, data, headers)

        return_value = self.trap.get()
        self.assertTrue(return_value)

    def test_pagerduty_202(self):
        """pagerduty 202
        """

        self.notify(self._http_post_202)
        results = self.trap.get(timeout=1)
        self.pagerduty_http_error(results, "202")

        return_value = self.trap.get()
        self.assertFalse(return_value)

    def test_pagerduty_400(self):
        """pagerduty 400
        """

        self.notify(self._http_post_400)
        results = self.trap.get(timeout=1)
        self.pagerduty_http_error(results, "400")

        return_value = self.trap.get()
        self.assertFalse(return_value)

    def test_pagerduty_403(self):
        """pagerduty 403
        """

        self.notify(self._http_post_403)
        results = self.trap.get(timeout=1)
        self.pagerduty_http_error(results, "403")

        return_value = self.trap.get()
        self.assertFalse(return_value)

    def test_pagerduty_500(self):
        """pagerduty 500
        """

        self.notify(self._http_post_500)
        results = self.trap.get(timeout=1)
        self.pagerduty_http_error(results, "500")

        return_value = self.trap.get()
        self.assertFalse(return_value)

    def test_pagerduty_504(self):
        """pagerduty 504
        """

        self.notify(self._http_post_504)
        results = self.trap.get(timeout=1)
        self.pagerduty_http_error(results, "504")

        return_value = self.trap.get()
        self.assertFalse(return_value)

    def test_pagerduty_exception(self):
        """pagerduty exception
        """

        self.notify(self._http_post_exception)

        results = self.trap.get(timeout=1)

        self.assertEqual(results, "timeout 50")

        results = self.trap.get(timeout=1)

        self.assertRegexpMatches(results, "Exception on pagerduty request")
        self.assertRegexpMatches(results, "key=<ABCDEF>")

        self.assertRaises(requests.exceptions.Timeout)

        return_value = self.trap.get()
        self.assertFalse(return_value)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import time

from monasca_common.kafka import consumer
from monasca_common.kafka import producer
from monasca_notification.common.utils import construct_notification_object
from monasca_notification.common.utils import get_db_repo
from monasca_notification.common.utils import get_statsd_client
from monasca_notification.processors import notification_processor

log = logging.getLogger(__name__)


class RetryEngine(object):
    def __init__(self, config):
        self._retry_interval = config['retry']['interval']
        self._retry_max = config['retry']['max_attempts']

        self._topics = {}
        self._topics['notification_topic'] = config['kafka']['notification_topic']
        self._topics['retry_topic'] = config['kafka']['notification_retry_topic']

        self._statsd = get_statsd_client(config)

        self._consumer = consumer.KafkaConsumer(
            config['kafka']['url'],
            config['zookeeper']['url'],
            config['zookeeper']['notification_retry_path'],
            config['kafka']['group'],
            config['kafka']['notification_retry_topic'])

        self._producer = producer.KafkaProducer(config['kafka']['url'])

        self._notifier = notification_processor.NotificationProcessor(config)
        self._db_repo = get_db_repo(config)

    def run(self):
        for raw_notification in self._consumer:
            message = raw_notification[1].message.value

            notification_data = json.loads(message)

            notification = construct_notification_object(self._db_repo, notification_data)

            if notification is None:
                self._consumer.commit()
                continue

            wait_duration = self._retry_interval - (
                time.time() - notification_data['notification_timestamp'])

            if wait_duration > 0:
                time.sleep(wait_duration)

            sent, failed = self._notifier.send([notification])

            if sent:
                self._producer.publish(self._topics['notification_topic'],
                                       [notification.to_json()])

            if failed:
                notification.retry_count += 1
                notification.notification_timestamp = time.time()
                if notification.retry_count < self._retry_max:
                    log.error(u"retry failed for {} with name {} "
                              u"at {}.  "
                              u"Saving for later retry.".format(notification.type,
                                                                notification.name,
                                                                notification.address))
                    self._producer.publish(self._topics['retry_topic'],
                                           [notification.to_json()])
                else:
                    log.error(u"retry failed for {} with name {} "
                              u"at {} after {} retries.  "
                              u"Giving up on retry."
                              .format(notification.type,
                                      notification.name,
                                      notification.address,
                                      self._retry_max))

            self._consumer.commit()



# Copyright 2015-2017 FUJITSU LIMITED
# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import logging
import pymysql

from monasca_notification.common.repositories.base import base_repo
from monasca_notification.common.repositories import exceptions as exc


log = logging.getLogger(__name__)


class MysqlRepo(base_repo.BaseRepo):
    def __init__(self, config):
        super(MysqlRepo, self).__init__(config)
        if 'ssl' in config['mysql']:
            self._mysql_ssl = config['mysql']['ssl']
        else:
            self._mysql_ssl = None

        if 'port' in config['mysql']:
            self._mysql_port = config['mysql']['port']
        else:
            #
            # If port isn't specified in the config file,
            # set it to the default value.
            #
            self._mysql_port = 3306

        self._mysql_host = config['mysql']['host']
        self._mysql_user = config['mysql']['user']
        self._mysql_passwd = config['mysql']['passwd']
        self._mysql_dbname = config['mysql']['db']
        self._mysql = None

    def _connect_to_mysql(self):
        self._mysql = None
        try:
            self._mysql = pymysql.connect(host=self._mysql_host,
                                          port=self._mysql_port,
                                          user=self._mysql_user,
                                          passwd=self._mysql_passwd,
                                          db=self._mysql_dbname,
                                          ssl=self._mysql_ssl,
                                          use_unicode=True,
                                          charset="utf8")
            self._mysql.autocommit(True)
        except pymysql.Error as e:
            log.exception('MySQL connect failed %s', e)
            raise

    def fetch_notifications(self, alarm):
        try:
            if self._mysql is None:
                self._connect_to_mysql()
            cur = self._mysql.cursor()
            cur.execute(self._find_alarm_action_sql, (alarm['alarmDefinitionId'], alarm['newState']))

            for row in cur:
                yield (row[0], row[1].lower(), row[2], row[3], row[4])
        except pymysql.Error as e:
            self._mysql = None
            log.exception("Couldn't fetch alarms actions %s", e)
            raise exc.DatabaseException(e)

    def get_alarm_current_state(self, alarm_id):
        try:
            if self._mysql is None:
                self._connect_to_mysql()
            cur = self._mysql.cursor()
            cur.execute(self._find_alarm_state_sql, alarm_id)
            row = cur.fetchone()
            state = row[0] if row is not None else None
            return state
        except pymysql.Error as e:
            self._mysql = None
            log.exception("Couldn't fetch the current alarm state %s", e)
            raise exc.DatabaseException(e)

    def fetch_notification_method_types(self):
        try:
            if self._mysql is None:
                self._connect_to_mysql()
            cur = self._mysql.cursor()
            cur.execute(self._find_all_notification_types_sql)

            for row in cur:
                yield (row[0])
        except pymysql.Error as e:
            self._mysql = None
            log.exception("Couldn't fetch notification types %s", e)
            raise exc.DatabaseException(e)

    def insert_notification_method_types(self, notification_types):
        try:
            if self._mysql is None:
                self._connect_to_mysql()
            cur = self._mysql.cursor()
            cur.executemany(self._insert_notification_types_sql, notification_types)

        except pymysql.IntegrityError as ignoredException:
            # If multiple instances of the notification engine tries to write the
            # same content at the same time, only one of them will succeed and others will
            # get duplicate primary key, integrity error. We can safely ignore this error.
            # This may happen only during the first start when the tables are empty.
            code, mesg = ignoredException.args
            if code == pymysql.constants.ER.DUP_ENTRY:
                log.debug("Notification type exists in DB. Ignoring the exception  {}".format(mesg))
            else:
                raise exc.DatabaseException(ignoredException)
        except pymysql.Error as e:
            self._mysql = None
            log.exception("Couldn't insert notification types %s", e)
            raise exc.DatabaseException(e)

    def get_notification(self, notification_id):
        try:
            if self._mysql is None:
                self._connect_to_mysql()
            cur = self._mysql.cursor()
            cur.execute(self._get_notification_sql, notification_id)
            row = cur.fetchone()
            if row is None:
                return None
            else:
                return [row[0], row[1].lower(), row[2], row[3]]
        except pymysql.Error as e:
            self._mysql = None
            log.exception("Couldn't fetch the notification method %s", e)
            raise exc.DatabaseException(e)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import email.header
import email.mime.text
import email.utils
import six
import smtplib
import time


from monasca_notification.plugins import abstract_notifier

EMAIL_SINGLE_HOST_BASE = u'''On host "{hostname}" for target "{target_host}" {message}

Alarm "{alarm_name}" transitioned to the {state} state at {timestamp} UTC
alarm_id: {alarm_id}

Lifecycle state: {lifecycle_state}
Link: {link}
Link to Grafana: {grafana_url}

With dimensions:
{metric_dimensions}'''

EMAIL_MULTIPLE_HOST_BASE = u'''On host "{hostname}" {message}

Alarm "{alarm_name}" transitioned to the {state} state at {timestamp} UTC
alarm_id: {alarm_id}

Lifecycle state: {lifecycle_state}
Link: {link}
Link to Grafana: {grafana_url}

With dimensions:
{metric_dimensions}'''

EMAIL_NO_HOST_BASE = u'''On multiple hosts {message}

Alarm "{alarm_name}" transitioned to the {state} state at {timestamp} UTC
Alarm_id: {alarm_id}

Lifecycle state: {lifecycle_state}
Link: {link}
Link to Grafana: {grafana_url}

With dimensions
{metric_dimensions}'''


class EmailNotifier(abstract_notifier.AbstractNotifier):

    def __init__(self, log):
        super(EmailNotifier, self).__init__()
        self._log = log
        self._smtp = None
        self._config = None

    def config(self, config):
        self._config = config
        self._smtp_connect()

    @property
    def type(self):
        return "email"

    @property
    def statsd_name(self):
        return "sent_smtp_count"

    def send_notification(self, notification):
        """Send the notification via email
             Returns the True upon success, False upon failure
        """

        # Get the "hostname" from the notification metrics if there is one
        hostname = []
        targethost = []

        for metric in notification.metrics:
            dimap = metric['dimensions']

            if 'hostname' in dimap and not dimap['hostname'] in hostname:
                hostname.append(dimap['hostname'])
            if 'target_host' in dimap and not dimap['target_host'] in targethost:
                targethost.append(dimap['target_host'])

        # Generate the message
        msg = self._create_msg(hostname, notification, targethost)

        if not self._smtp and not self._smtp_connect():
            return False

        try:
            self._sendmail(notification, msg)
            return True
        except smtplib.SMTPServerDisconnected:
            self._log.warn('SMTP server disconnected. '
                           'Will reconnect and retry message.')
            self._smtp_connect()
        except smtplib.SMTPException:
            self._email_error(notification)
            return False

        try:
            self._sendmail(notification, msg)
            return True
        except smtplib.SMTPException:
            self._email_error(notification)
            return False

    def _sendmail(self, notification, msg):
        self._smtp.sendmail(self._config['from_addr'],
                            notification.address,
                            msg.as_string())
        self._log.debug("Sent email to {}, notification {}".format(notification.address,
                                                                   notification.to_json()))

    def _email_error(self, notification):
        self._log.exception("Error sending Email Notification")
        self._log.error("Failed email: {}".format(notification.to_json()))

    def _smtp_connect(self):
        """Connect to the smtp server
        """
        self._log.info("Connecting to Email Server {}".format(self._config['server']))

        try:
            smtp = smtplib.SMTP(self._config['server'],
                                self._config['port'],
                                timeout=self._config['timeout'])

            if ('user', 'password') in self._config.keys():
                smtp.login(self._config['user'], self._config['password'])

            self._smtp = smtp
            return True
        except Exception:
            self._log.exception("Unable to connect to email server.")
            return False

    def _create_msg(self, hostname, notification, targethost=None):
        """Create two kind of messages:
        1. Notifications that include metrics with a hostname as a dimension. There may be more than one hostname.
           We will only report the hostname if there is only one.
        2. Notifications that do not include metrics and therefore no hostname. Example: API initiated changes.
           * A third notification type which include metrics but do not include a hostname will
           be treated as type #2.
        """
        timestamp = time.asctime(time.gmtime(notification.alarm_timestamp))

        alarm_seconds = notification.alarm_timestamp
        alarm_ms = int(round(alarm_seconds * 1000))

        graf_url = self._get_link_url(notification.metrics[0], alarm_ms)

        dimensions = _format_dimensions(notification)

        if len(hostname) == 1:  # Type 1
            if targethost:
                text = EMAIL_SINGLE_HOST_BASE.format(
                    hostname=hostname[0],
                    target_host=targethost[0],
                    message=notification.message.lower(),
                    alarm_name=notification.alarm_name,
                    state=notification.state,
                    timestamp=timestamp,
                    alarm_id=notification.alarm_id,
                    metric_dimensions=dimensions,
                    link=notification.link,
                    grafana_url=graf_url,
                    lifecycle_state=notification.lifecycle_state
                )
                subject = u'{} {} "{}" for Host: {} Target: {}'.format(
                    notification.state, notification.severity,
                    notification.alarm_name, hostname[0],
                    targethost[0]
                )
            else:
                text = EMAIL_MULTIPLE_HOST_BASE.format(
                    hostname=hostname[0],
                    message=notification.message.lower(),
                    alarm_name=notification.alarm_name,
                    state=notification.state,
                    timestamp=timestamp,
                    alarm_id=notification.alarm_id,
                    metric_dimensions=dimensions,
                    link=notification.link,
                    grafana_url=graf_url,
                    lifecycle_state=notification.lifecycle_state
                )
                subject = u'{} {} "{}" for Host: {}'.format(
                    notification.state, notification.severity,
                    notification.alarm_name, hostname[0])
        else:  # Type 2
            text = EMAIL_NO_HOST_BASE.format(
                message=notification.message.lower(),
                alarm_name=notification.alarm_name,
                state=notification.state,
                timestamp=timestamp,
                alarm_id=notification.alarm_id,
                metric_dimensions=dimensions,
                link=notification.link,
                grafana_url=graf_url,
                lifecycle_state=notification.lifecycle_state
            )
            subject = u'{} {} "{}" '.format(notification.state,
                                            notification.severity,
                                            notification.alarm_name)

        msg = email.mime.text.MIMEText(text, 'plain', 'utf-8')
        msg['Subject'] = email.header.Header(subject, 'utf-8')
        msg['From'] = self._config['from_addr']
        msg['To'] = notification.address
        msg['Date'] = email.utils.formatdate(localtime=True, usegmt=True)

        return msg

    def _get_link_url(self, metric, timestamp_ms):
        """Returns the url to Grafana including a query with the
        respective metric info (name, dimensions, timestamp)
        :param metric: the metric for which to display the graph in Grafana
        :param timestamp_ms: timestamp of the alarm for the metric in milliseconds
        :return: the url to the graph for the given metric or None if no Grafana host
        has been defined.
        """

        grafana_url = self._config.get('grafana_url', None)
        if grafana_url is None:
            return None

        url = ''
        metric_query = ''

        metric_query = "?metric=%s" % metric['name']

        dimensions = metric['dimensions']
        for key, value in six.iteritems(dimensions):
            metric_query += "&dim_%s=%s" % (key, value)

        # Show the graph within a range of ten minutes before and after the alarm occurred.
        offset = 600000
        from_ms = timestamp_ms - offset
        to_ms = timestamp_ms + offset
        time_query = "&from=%s&to=%s" % (from_ms, to_ms)

        url = grafana_url + '/dashboard/script/drilldown.js'

        return url + metric_query + time_query


def _format_dimensions(notification):
    dimension_sets = []
    for metric in notification.metrics:
        dimension_sets.append(metric['dimensions'])

    dim_set_strings = []
    for dimension_set in dimension_sets:
        key_value_pairs = []
        for key, value in dimension_set.items():
            key_value_pairs.append(u'    {}: {}'.format(key, value))

        set_string = u'  {\n' + u',\n'.join(key_value_pairs) + u'\n  }'
        dim_set_strings.append(set_string)

    dimensions = u'[\n' + u',\n'.join(dim_set_strings) + u' \n]'

    return dimensions



# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the AlarmProcessor"""

import collections
import json
import mock
import time
import unittest

from monasca_notification import notification as m_notification
from monasca_notification.processors import alarm_processor

alarm_tuple = collections.namedtuple('alarm_tuple', ['offset', 'message'])
message_tuple = collections.namedtuple('message_tuple', ['value'])


class TestAlarmProcessor(unittest.TestCase):
    def setUp(self):
        self.trap = []

    def _create_raw_alarm(self, partition, offset, message):
        """Create a raw alarm, with the given message dictionary.
        """
        json_msg = json.dumps({'alarm-transitioned': message})
        msg_tuple = message_tuple(json_msg)
        return [partition, alarm_tuple(offset, msg_tuple)]

    @mock.patch('pymysql.connect')
    @mock.patch('monasca_notification.processors.alarm_processor.log')
    def _run_alarm_processor(self, alarm, sql_response, mock_log, mock_mysql):
        """Runs a mocked alarm processor reading from queue while running, returns (queue_message, log_message)
        """
        # Since the log runs in another thread I can mock it directly, instead change the methods to put to a queue
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.exception = self.trap.append

        # Setup the sql response
        if sql_response is not None:
            mock_mysql.return_value = mock_mysql
            mock_mysql.cursor.return_value = mock_mysql
            mock_mysql.__iter__.return_value = sql_response

        config = {'mysql': {'ssl': None,
                            'host': 'mysql_host',
                            'port': 'mysql_port',
                            'user': 'mysql_user',
                            'db': 'dbname',
                            'passwd': 'mysql_passwd'},
                  'statsd': {'host': 'localhost',
                             'port': 8125}}

        processor = alarm_processor.AlarmProcessor(600, config)

        return processor.to_notification(alarm)

    def test_invalid_alarm(self):
        """Invalid Alarms, should log and error and push to the finished queue."""
        alarm = self._create_raw_alarm(0, 1, {'invalid': 'invalid_alarm'})
        notifications, partition, offset = self._run_alarm_processor(alarm, None)
        self.assertEqual(notifications, [])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 1)

        invalid_msg = ('Invalid Alarm format skipping partition 0, offset 1\n'
                       'ErrorAlarm data missing field actionsEnabled')

        self.assertIn(invalid_msg, self.trap)

    def test_old_timestamp(self):
        """Should cause the alarm_ttl to fire log a warning and push to finished queue."""
        timestamp = 1375346830042
        alarm_dict = {"tenantId": "0", "alarmDefinitionId": "0", "alarmId": "1", "alarmName": "test Alarm",
                      "oldState": "OK", "newState": "ALARM", "stateChangeReason": "I am alarming!",
                      "timestamp": timestamp, "actionsEnabled": 1, "metrics": "cpu_util",
                      "severity": "LOW", "link": "http://some-place.com", "lifecycleState": "OPEN"}
        alarm = self._create_raw_alarm(0, 2, alarm_dict)
        expected_datetime = time.ctime(timestamp / 1000)

        notifications, partition, offset = self._run_alarm_processor(alarm, None)

        self.assertEqual(notifications, [])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 2)

        old_msg = ('Received alarm older than the ttl, skipping. '
                   'Alarm from {datetime}'.format(datetime=expected_datetime))

        self.assertIn(old_msg, self.trap)

    def test_no_notifications(self):
        """Test an alarm with no defined notifications
        """
        alarm_dict = {"tenantId": "0", "alarmDefinitionId": "0", "alarmId": "1", "alarmName": "test Alarm",
                      "oldState": "OK", "newState": "ALARM", "stateChangeReason": "I am alarming!",
                      "timestamp": time.time() * 1000, "actionsEnabled": 1, "metrics": "cpu_util",
                      "severity": "LOW", "link": "http://some-place.com", "lifecycleState": "OPEN"}
        alarm = self._create_raw_alarm(0, 3, alarm_dict)

        notifications, partition, offset = self._run_alarm_processor(alarm, None)

        self.assertEqual(notifications, [])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 3)

    def test_valid_notification(self):
        """Test a valid notification, being put onto the notification_queue
        """
        alarm_dict = {"tenantId": "0", "alarmDefinitionId": "0", "alarmId": "1", "alarmName": "test Alarm",
                      "oldState": "OK", "newState": "ALARM", "stateChangeReason": "I am alarming!",
                      "timestamp": time.time() * 1000, "actionsEnabled": 1, "metrics": "cpu_util",
                      "severity": "LOW", "link": "http://some-place.com", "lifecycleState": "OPEN"}
        alarm = self._create_raw_alarm(0, 4, alarm_dict)

        sql_response = [[1, 'EMAIL', 'test notification', 'me@here.com', 0]]
        notifications, partition, offset = self._run_alarm_processor(alarm, sql_response)

        test_notification = m_notification.Notification(1, 'email', 'test notification',
                                                        'me@here.com', 0, 0, alarm_dict)

        self.assertEqual(notifications, [test_notification])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 4)

    def test_two_valid_notifications(self):
        alarm_dict = {"tenantId": "0", "alarmDefinitionId": "0", "alarmId": "1", "alarmName": "test Alarm",
                      "oldState": "OK", "newState": "ALARM", "stateChangeReason": "I am alarming!",
                      "timestamp": time.time() * 1000, "actionsEnabled": 1, "metrics": "cpu_util",
                      "severity": "LOW", "link": "http://some-place.com", "lifecycleState": "OPEN"}

        alarm = self._create_raw_alarm(0, 5, alarm_dict)

        sql_response = [[1, 'EMAIL', 'test notification', 'me@here.com', 0],
                        [2, 'EMAIL', 'test notification2', 'me@here.com', 0]]
        notifications, partition, offset = self._run_alarm_processor(alarm, sql_response)

        test_notification = m_notification.Notification(1, 'email', 'test notification',
                                                        'me@here.com', 0, 0, alarm_dict)
        test_notification2 = m_notification.Notification(2, 'email', 'test notification2',
                                                         'me@here.com', 0, 0, alarm_dict)

        self.assertEqual(notifications, [test_notification, test_notification2])
        self.assertEqual(partition, 0)
        self.assertEqual(offset, 5)



# Copyright 2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from mock import patch
import unittest

from monasca_notification.common import utils


class TestStatsdConnection(unittest.TestCase):
    extra_dimensions = {'foo': 'bar'}
    base_name = 'monasca'

    def test_statsd_default_connection(self):
        config = {}
        with patch(
                'monasca_notification.common.utils.monascastatsd.Client') as c:
            utils.get_statsd_client(config)
            c.assert_called_once_with(dimensions=utils.NOTIFICATION_DIMENSIONS,
                                      name=self.base_name)

    def test_statsd_config_connection(self):
        port_number = 9999
        hostname = 'www.example.org'
        config = {'statsd': {'host': hostname, 'port': port_number}}
        with patch(
                'monasca_notification.common.utils.monascastatsd.Client') as c:
            utils.get_statsd_client(config)
            c.assert_called_once_with(dimensions=utils.NOTIFICATION_DIMENSIONS,
                                      name=self.base_name,
                                      port=port_number,
                                      host=hostname)

    def test_statsd_update_dimmensions(self):
        config = {}
        expected_dimensions = utils.NOTIFICATION_DIMENSIONS.copy()
        expected_dimensions.update(self.extra_dimensions)
        with patch(
                'monasca_notification.common.utils.monascastatsd.Client') as c:
            utils.get_statsd_client(config, dimensions=self.extra_dimensions)
            c.assert_called_once_with(dimensions=expected_dimensions,
                                      name=self.base_name)



# (C) Copyright 2014-2017 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Notification Engine
    This engine reads alarms from Kafka and then notifies the customer using their configured notification method.
"""

import logging
import logging.config
import multiprocessing
import os
import signal
import sys
import time
import yaml

from monasca_notification.notification_engine import NotificationEngine
from monasca_notification.periodic_engine import PeriodicEngine
from monasca_notification.retry_engine import RetryEngine

log = logging.getLogger(__name__)
processors = []  # global list to facilitate clean signal handling
exiting = False


def clean_exit(signum, frame=None):
    """Exit all processes attempting to finish uncommitted active work before exit.
         Can be called on an os signal or no zookeeper losing connection.
    """
    global exiting
    if exiting:
        # Since this is set up as a handler for SIGCHLD when this kills one
        # child it gets another signal, the global exiting avoids this running
        # multiple times.
        log.debug('Exit in progress clean_exit received additional signal %s' % signum)
        return

    log.info('Received signal %s, beginning graceful shutdown.' % signum)
    exiting = True
    wait_for_exit = False

    for process in processors:
        try:
            if process.is_alive():
                process.terminate()  # Sends sigterm which any processes after a notification is sent attempt to handle
                wait_for_exit = True
        except Exception:  # nosec
            # There is really nothing to do if the kill fails, so just go on.
            # The # nosec keeps bandit from reporting this as a security issue
            pass

    # wait for a couple seconds to give the subprocesses a chance to shut down correctly.
    if wait_for_exit:
        time.sleep(2)

    # Kill everything, that didn't already die
    for child in multiprocessing.active_children():
        log.debug('Killing pid %s' % child.pid)
        try:
            os.kill(child.pid, signal.SIGKILL)
        except Exception:  # nosec
            # There is really nothing to do if the kill fails, so just go on.
            # The # nosec keeps bandit from reporting this as a security issue
            pass

    if signum == signal.SIGTERM:
        sys.exit(0)

    sys.exit(signum)


def start_process(process_type, config, *args):
    log.info("start process: {}".format(process_type))
    p = process_type(config, *args)
    p.run()


def main(argv=None):
    if argv is None:
        argv = sys.argv
    if len(argv) == 2:
        config_file = argv[1]
    elif len(argv) > 2:
        print("Usage: " + argv[0] + " <config_file>")
        print("Config file defaults to /etc/monasca/notification.yaml")
        return 1
    else:
        config_file = '/etc/monasca/notification.yaml'

    config = yaml.safe_load(open(config_file, 'rb'))

    # Setup logging
    try:
        if config['logging']['raise_exceptions'] is True:
            logging.raiseExceptions = True
        else:
            logging.raiseExceptions = False
    except KeyError:
        logging.raiseExceptions = False
        pass
    logging.config.dictConfig(config['logging'])

    for proc in range(0, config['processors']['notification']['number']):
        processors.append(multiprocessing.Process(
            target=start_process, args=(NotificationEngine, config)))

    processors.append(multiprocessing.Process(
        target=start_process, args=(RetryEngine, config)))

    if 60 in config['kafka']['periodic']:
        processors.append(multiprocessing.Process(
            target=start_process, args=(PeriodicEngine, config, 60)))

    try:
        log.info('Starting processes')
        for process in processors:
            process.start()

        # The signal handlers must be added after the processes start otherwise
        # they run on all processes
        signal.signal(signal.SIGCHLD, clean_exit)
        signal.signal(signal.SIGINT, clean_exit)
        signal.signal(signal.SIGTERM, clean_exit)

        while True:
            time.sleep(10)

    except Exception:
        log.exception('Error! Exiting.')
        clean_exit(signal.SIGKILL)

if __name__ == "__main__":
    sys.exit(main())



# Copyright 2015 FUJITSU LIMITED
# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import logging
from sqlalchemy import engine_from_config, MetaData
from sqlalchemy.sql import select, bindparam, and_, insert
from sqlalchemy.exc import DatabaseError

from monasca_notification.common.repositories import exceptions as exc
from monasca_notification.common.repositories.orm import models

log = logging.getLogger(__name__)


class OrmRepo(object):
    def __init__(self, config):
        self._orm_engine = engine_from_config(config['database']['orm'], prefix='')

        metadata = MetaData()

        aa = models.create_alarm_action_model(metadata).alias('aa')
        nm = models.create_notification_method_model(metadata).alias('nm')
        nmt_insert = models.create_notification_method_type_model(metadata)
        nmt = nmt_insert.alias('nmt')
        a = models.create_alarm_model(metadata).alias('a')

        self._orm_query = select([nm.c.id, nm.c.type, nm.c.name, nm.c.address, nm.c.period])\
            .select_from(aa.join(nm, aa.c.action_id == nm.c.id))\
            .where(
                and_(aa.c.alarm_definition_id == bindparam('alarm_definition_id'),
                     aa.c.alarm_state == bindparam('alarm_state')))

        self._orm_get_alarm_state = select([a.c.state]).where(a.c.id == bindparam('alarm_id'))

        self._orm_nmt_query = select([nmt.c.name])

        self._orm_get_notification = select([nm.c.name, nm.c.type, nm.c.address, nm.c.period])\
            .where(nm.c.id == bindparam('notification_id'))

        self._orm_add_notification_type = insert(nmt_insert).values(name=bindparam('b_name'))

        self._orm = None

    def fetch_notifications(self, alarm):
        try:
            with self._orm_engine.connect() as conn:
                log.debug('Orm query {%s}', str(self._orm_query))
                notifications = conn.execute(self._orm_query,
                                             alarm_definition_id=alarm['alarmDefinitionId'],
                                             alarm_state=alarm['newState'])

                return [(row[0], row[1].lower(), row[2], row[3], row[4]) for row in notifications]
        except DatabaseError as e:
            log.exception("Couldn't fetch alarms actions %s", e)
            raise exc.DatabaseException(e)

    def get_alarm_current_state(self, alarm_id):
        try:
            with self._orm_engine.connect() as conn:
                log.debug('Orm query {%s}', str(self._orm_get_alarm_state))
                result = conn.execute(self._orm_get_alarm_state,
                                      alarm_id=alarm_id)
                row = result.fetchone()
                state = row[0] if row is not None else None
                return state
        except DatabaseError as e:
            log.exception("Couldn't fetch the current alarm state %s", e)
            raise exc.DatabaseException(e)

    def fetch_notification_method_types(self):
        try:
            with self._orm_engine.connect() as conn:
                log.debug('Orm query {%s}', str(self._orm_nmt_query))
                notification_method_types = conn.execute(self._orm_nmt_query).fetchall()

                return [row[0] for row in notification_method_types]
        except DatabaseError as e:
            log.exception("Couldn't fetch notification method types %s", e)
            raise exc.DatabaseException(e)

    def insert_notification_method_types(self, notification_types):
        # This function can be called by multiple processes at same time.
        # In that case, only the first one will succeed and the others will fail.
        # We can ignore this error when the types have already been registered.
        try:
            with self._orm_engine.connect() as conn:
                for notification_type in notification_types:
                    conn.execute(self._orm_add_notification_type, b_name=notification_type)

        except DatabaseError as e:
            log.debug("Failed to insert notification types %s", notification_types)
            raise exc.DatabaseException(e)

    def get_notification(self, notification_id):
        try:
            with self._orm_engine.connect() as conn:
                log.debug('Orm query {%s}', str(self._orm_get_notification))
                result = conn.execute(self._orm_get_notification,
                                      notification_id=notification_id)
                notification = result.fetchone()
                if notification is None:
                    return None
                else:
                    return [notification[0], notification[1].lower(), notification[2], notification[3]]
        except DatabaseError as e:
            log.exception("Couldn't fetch the notification method %s", e)
            raise exc.DatabaseException(e)



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import time

from monasca_common.kafka import consumer
from monasca_common.kafka import producer
from monasca_notification.common.repositories import exceptions
from monasca_notification.common.utils import construct_notification_object
from monasca_notification.common.utils import get_db_repo
from monasca_notification.common.utils import get_statsd_client
from monasca_notification.processors import notification_processor

log = logging.getLogger(__name__)


class PeriodicEngine(object):
    def __init__(self, config, period):
        self._topic_name = config['kafka']['periodic'][period]

        self._statsd = get_statsd_client(config)

        zookeeper_path = config['zookeeper']['periodic_path'][period]
        self._consumer = consumer.KafkaConsumer(config['kafka']['url'],
                                                config['zookeeper']['url'],
                                                zookeeper_path,
                                                config['kafka']['group'],
                                                self._topic_name)

        self._producer = producer.KafkaProducer(config['kafka']['url'])

        self._notifier = notification_processor.NotificationProcessor(config)
        self._db_repo = get_db_repo(config)
        self._period = period

    def _keep_sending(self, alarm_id, original_state, type, period):
        try:
            current_state = self._db_repo.get_alarm_current_state(alarm_id)
        except exceptions.DatabaseException:
            log.debug('Database Error.  Attempting reconnect')
            current_state = self._db_repo.get_alarm_current_state(alarm_id)

        # Alarm was deleted
        if current_state is None:
            return False
        # Alarm state changed
        if current_state != original_state:
            return False
        # Period changed
        if period != self._period:
            return False
        if type != "webhook":
            return False

        return True

    def run(self):
        for raw_notification in self._consumer:
            message = raw_notification[1].message.value
            notification_data = json.loads(message)

            notification = construct_notification_object(self._db_repo, notification_data)

            if notification is None:
                self._consumer.commit()
                continue

            if self._keep_sending(notification.alarm_id,
                                  notification.state,
                                  notification.type,
                                  notification.period):

                wait_duration = notification.period - (
                    time.time() - notification_data['notification_timestamp'])

                if wait_duration > 0:
                    time.sleep(wait_duration)

                notification.notification_timestamp = time.time()

                self._notifier.send([notification])
                self._producer.publish(self._topic_name,
                                       [notification.to_json()])

            self._consumer.commit()



# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from monasca_notification.common.repositories import exceptions as exc
from monasca_notification.common.utils import get_db_repo
from monasca_notification.common.utils import get_statsd_client
from monasca_notification.types import notifiers

log = logging.getLogger(__name__)


class NotificationProcessor(object):

    def __init__(self, config):
        self.statsd = get_statsd_client(config)
        notifiers.init(self.statsd)
        notifiers.load_plugins(config['notification_types'])
        notifiers.config(config['notification_types'])
        self._db_repo = get_db_repo(config)
        self.insert_configured_plugins()

    def _remaining_plugin_types(self):
        configured_plugin_types = notifiers.enabled_notifications()
        persisted_plugin_types = self._db_repo.fetch_notification_method_types()

        return set(configured_plugin_types) - set(persisted_plugin_types)

    def insert_configured_plugins(self):
        """Persists configured plugin types in DB
             For each notification type configured add it in db, if it is not there
        """
        remaining_plugin_types = self._remaining_plugin_types()

        max_retry = len(remaining_plugin_types)
        retry_count = 0
        while remaining_plugin_types:
            retry_count = retry_count + 1
            try:
                log.info("New plugins detected: Adding new notification types %s to database",
                         remaining_plugin_types)
                self._db_repo.insert_notification_method_types(remaining_plugin_types)
            except exc.DatabaseException as e:
                # There is a possibility the other process has already registered the type.
                remaining_plugin_types = self._remaining_plugin_types()
                if remaining_plugin_types and (retry_count >= max_retry):
                    log.exception("Couldn't insert notification types %s", e)
                    raise e
                else:
                    log.debug("Plugin already exists. Ignore exception")

    def send(self, notifications):
        """Send the notifications
             For each notification in a message it is sent according to its type.
             If all notifications fail the alarm partition/offset are added to the finished queue
        """

        invalid_type_count = self.statsd.get_counter(name='invalid_type_count')
        sent_failed_count = self.statsd.get_counter(name='sent_failed_count')

        sent, failed, invalid = notifiers.send_notifications(notifications)

        if failed:
            sent_failed_count.increment(len(failed))

        if invalid:
            invalid_type_count.increment(len(invalid))

        return sent, failed



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import ujson as json

from monasca_notification.plugins import abstract_notifier


class WebhookNotifier(abstract_notifier.AbstractNotifier):
    def __init__(self, log):
        self._log = log

    def config(self, config_dict):
        self._config = {'timeout': 5}
        self._config.update(config_dict)

    @property
    def type(self):
        return "webhook"

    @property
    def statsd_name(self):
        return 'sent_webhook_count'

    def send_notification(self, notification):
        """Send the notification via webhook
            Posts on the given url
        """

        body = {'alarm_id': notification.alarm_id,
                'alarm_definition_id': notification.raw_alarm['alarmDefinitionId'],
                'alarm_name': notification.alarm_name,
                'alarm_description': notification.raw_alarm['alarmDescription'],
                'alarm_timestamp': notification.alarm_timestamp,
                'state': notification.state,
                'old_state': notification.raw_alarm['oldState'],
                'message': notification.message,
                'tenant_id': notification.tenant_id,
                'metrics': notification.metrics}

        headers = {'content-type': 'application/json'}

        url = notification.address

        try:
            # Posting on the given URL
            result = requests.post(url=url,
                                   data=json.dumps(body),
                                   headers=headers,
                                   timeout=self._config['timeout'])

            if result.status_code in range(200, 300):
                self._log.info("Notification successfully posted.")
                return True
            else:
                self._log.error("Received an HTTP code {} when trying to post on URL {}."
                                .format(result.status_code, url))
                return False
        except Exception:
            self._log.exception("Error trying to post on URL {}".format(url))
            return False



# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json


class Notification(object):
    """An abstract base class used to define the notification interface
       and common functions
    """
    __slots__ = (
        'address',
        'alarm_id',
        'alarm_name',
        'alarm_timestamp',
        'id',
        'message',
        'name',
        'notification_timestamp',
        'state',
        'severity',
        'link',
        'lifecycle_state',
        'tenant_id',
        'type',
        'metrics',
        'retry_count',
        'raw_alarm',
        'period',
        'periodic_topic'
    )

    def __init__(self, id, type, name, address, period, retry_count, alarm):
        """Setup the notification object
             id - The notification id
             type - The notification type
             name - Name used in sending
             address - where to send the notification
             period - period of sending the notificationv
             retry_count - number of times we've tried to send
             alarm - info that caused the notification
             notifications that come after this one to remain uncommitted.
             Note that data may include six.text_type strings.
        """
        self.id = id
        self.address = address
        self.name = name
        self.type = type
        self.retry_count = retry_count

        self.raw_alarm = alarm

        self.alarm_id = alarm['alarmId']
        self.alarm_name = alarm['alarmName']
        # The event timestamp is in milliseconds
        self.alarm_timestamp = int(alarm['timestamp'] / 1000.0)
        self.message = alarm['stateChangeReason']
        self.state = alarm['newState']
        self.severity = alarm['severity']
        self.link = alarm['link']
        self.lifecycle_state = alarm['lifecycleState']
        self.tenant_id = alarm['tenantId']
        self.metrics = alarm['metrics']

        # to be updated on actual notification send time
        self.notification_timestamp = None

        # set periodic topic
        self.periodic_topic = period
        self.period = period

    def __eq__(self, other):
        if not isinstance(other, Notification):
            return False

        for attrib in self.__slots__:
            if not getattr(self, attrib) == getattr(other, attrib):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_json(self):
        """Return json representation
        """
        notification_fields = [
            'id',
            'type',
            'name',
            'address',
            'retry_count',
            'raw_alarm',
            'alarm_id',
            'alarm_name',
            'alarm_timestamp',
            'message',
            'notification_timestamp',
            'state',
            'severity',
            'link',
            'lifecycle_state',
            'tenant_id',
            'period',
            'periodic_topic'
        ]
        notification_data = {name: getattr(self, name)
                             for name in notification_fields}
        return json.dumps(notification_data)



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

from monasca_common.simport import simport
from monasca_notification.plugins import email_notifier
from monasca_notification.plugins import pagerduty_notifier
from monasca_notification.plugins import webhook_notifier

log = logging.getLogger(__name__)

possible_notifiers = None
configured_notifiers = None
statsd_counter = None

statsd = None
statsd_timer = None


def init(statsd_obj):
    global statsd, statsd_timer, \
        possible_notifiers, configured_notifiers,\
        statsd_counter

    statsd = statsd_obj
    statsd_timer = statsd.get_timer()

    statsd_counter = {}
    configured_notifiers = {}

    possible_notifiers = [
        email_notifier.EmailNotifier(log),
        webhook_notifier.WebhookNotifier(log),
        pagerduty_notifier.PagerdutyNotifier(log)
    ]


def load_plugins(config):
    global possible_notifiers

    for plugin_class in config.get("plugins", []):
        try:
            possible_notifiers.append(simport.load(plugin_class)(log))
        except Exception:
            log.exception("unable to load the class {0} , ignoring it".format(plugin_class))


def enabled_notifications():
    global configured_notifiers

    results = []
    for key in configured_notifiers:
        results.append(key.upper())
    return results


def config(cfg):
    global possible_notifiers, configured_notifiers, statsd_counter

    formatted_config = {t.lower(): v for t, v in cfg.items()}
    for notifier in possible_notifiers:
        ntype = notifier.type.lower()
        if ntype in formatted_config:
            try:
                notifier.config(formatted_config[ntype])
                configured_notifiers[ntype] = notifier
                statsd_counter[ntype] = statsd.get_counter(notifier.statsd_name)
                log.info("{} notification ready".format(ntype))
            except Exception:
                log.exception("config exception for {}".format(ntype))
        else:
            log.warn("No config data for type: {}".format(ntype))
    config_with_no_notifiers = set(formatted_config.keys()) - set(configured_notifiers.keys())
    # Plugins section contains only additional plugins and should not be
    # considered as a separate plugin
    if 'plugins' in config_with_no_notifiers:
        config_with_no_notifiers.remove('plugins')
    if config_with_no_notifiers:
        log.warn("No notifiers found for {0}". format(", ".join(config_with_no_notifiers)))


def send_notifications(notifications):
    sent = []
    failed = []
    invalid = []

    for notification in notifications:
        ntype = notification.type
        if ntype not in configured_notifiers:
            log.warn("attempting to send unconfigured notification: {}".format(ntype))
            invalid.append(notification)
            continue

        notification.notification_timestamp = time.time()

        with statsd_timer.time(ntype + '_time'):
            result = send_single_notification(notification)

        if result:
            sent.append(notification)
            statsd_counter[ntype].increment(1)
        else:
            failed.append(notification)

    return sent, failed, invalid


def send_single_notification(notification):
    global configured_notifiers

    ntype = notification.type
    try:
        return configured_notifiers[ntype].send_notification(notification)
    except Exception:
        log.exception("send_notification exception for {}".format(ntype))
        return False



# (C) Copyright 2014-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests the notification class."""

import json
from monasca_notification import notification


def test_json():
    """Test the to_json method to verify it behaves as expected.
    """
    ts = 1429029121239
    alarm = {'alarmId': 'alarmId',
             'alarmName': 'alarmName',
             'timestamp': ts,
             'stateChangeReason': 'stateChangeReason',
             'newState': 'newState',
             'severity': 'LOW',
             "link": "some-link",
             "lifecycleState": "OPEN",
             'tenantId': 'tenantId',
             'metrics': 'cpu_util'}
    test_notification = notification.Notification(1, 'ntype', 'name',
                                                  'address', 0, 0, alarm)

    expected_dict = {u'id': 1,
                     u'name': u'name',
                     u'type': u'ntype',
                     u'notification_timestamp': None,
                     u'tenant_id': u'tenantId',
                     u'alarm_name': u'alarmName',
                     u'alarm_id': u'alarmId',
                     u'state': u'newState',
                     u'severity': u'LOW',
                     u'link': u'some-link',
                     u'lifecycle_state': u'OPEN',
                     u'alarm_timestamp': ts / 1000,
                     u'address': u'address',
                     u'message': u'stateChangeReason',
                     u'period': 0,
                     u'periodic_topic': 0,
                     u'retry_count': 0,
                     u'raw_alarm': {
                         u'alarmId': u'alarmId',
                         u'alarmName': u'alarmName',
                         u'timestamp': ts,
                         u'stateChangeReason': u'stateChangeReason',
                         u'newState': u'newState',
                         u'severity': u'LOW',
                         u'link': u'some-link',
                         u'lifecycleState': u'OPEN',
                         u'tenantId': u'tenantId',
                         u'metrics': u'cpu_util'}}

    # Compare as dicts so ordering is not an issue
    assert json.loads(test_notification.to_json()) == expected_dict


def test_json_non_zero_period():
    """Test the to_json method to verify it behaves as expected.
    """
    ts = 1429029121239
    alarm = {'alarmId': 'alarmId',
             'alarmName': 'alarmName',
             'timestamp': ts,
             'stateChangeReason': 'stateChangeReason',
             'newState': 'newState',
             'severity': 'LOW',
             "link": "some-link",
             "lifecycleState": "OPEN",
             'tenantId': 'tenantId',
             'metrics': 'cpu_util'}
    test_notification = notification.Notification(1, 'ntype', 'name',
                                                  'address', 60, 0, alarm)

    expected_dict = {u'id': 1,
                     u'name': u'name',
                     u'type': u'ntype',
                     u'notification_timestamp': None,
                     u'tenant_id': u'tenantId',
                     u'alarm_name': u'alarmName',
                     u'alarm_id': u'alarmId',
                     u'state': u'newState',
                     u'severity': u'LOW',
                     u'link': u'some-link',
                     u'lifecycle_state': u'OPEN',
                     u'alarm_timestamp': ts / 1000,
                     u'address': u'address',
                     u'message': u'stateChangeReason',
                     u'period': 60,
                     u'periodic_topic': 60,
                     u'retry_count': 0,
                     u'raw_alarm': {
                         u'alarmId': u'alarmId',
                         u'alarmName': u'alarmName',
                         u'timestamp': ts,
                         u'stateChangeReason': u'stateChangeReason',
                         u'newState': u'newState',
                         u'severity': u'LOW',
                         u'link': u'some-link',
                         u'lifecycleState': u'OPEN',
                         u'tenantId': u'tenantId',
                         u'metrics': u'cpu_util'}}

    # Compare as dicts so ordering is not an issue
    assert json.loads(test_notification.to_json()) == expected_dict


def test_equal():
    alarm = {'alarmId': 'alarmId',
             'alarmName': 'alarmName',
             'timestamp': 1429029121239,
             'stateChangeReason': 'stateChangeReason',
             'newState': 'newState',
             'severity': 'LOW',
             "link": "some-link",
             "lifecycleState": "OPEN",
             'tenantId': 'tenantId',
             'metrics': 'cpu_util'}
    test_notification = notification.Notification(0, 'ntype', 'name',
                                                  'address', 0, 0, alarm)
    test_notification2 = notification.Notification(0, 'ntype', 'name',
                                                   'address', 0, 0, alarm)

    assert(test_notification == test_notification2)


def test_unequal():
    alarm = {'alarmId': 'alarmId',
             'alarmName': 'alarmName',
             'timestamp': 1429029121239,
             'stateChangeReason': 'stateChangeReason',
             'newState': 'newState',
             'severity': 'LOW',
             "link": "some-link",
             "lifecycleState": "OPEN",
             'tenantId': 'tenantId',
             'metrics': 'cpu_util'}
    test_notification = notification.Notification(0, 'ntype', 'name',
                                                  'address', 0, 0, alarm)
    test_notification2 = notification.Notification(1, 'ntype', 'name',
                                                   'address', 0, 0, alarm)
    test_notification2.alarm_id = None

    assert(test_notification != test_notification2)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import mock
from oslotest import base
import six

from monasca_notification import notification as m_notification
from monasca_notification.plugins import jira_notifier

if six.PY2:
    import Queue as queue
else:
    import queue


def alarm():
    return {'tenantId': '0',
            'alarmId': '0',
            'alarmDefinitionId': 0,
            'alarmName': 'test Alarm',
            'alarmDescription': 'test Alarm description',
            'oldState': 'OK',
            'newState': 'ALARM',
            'severity': 'CRITICAL',
            'link': 'some-link',
            'lifecycleState': 'OPEN',
            'stateChangeReason': 'I am alarming!',
            'timestamp': 1429023453632,
            'metrics': {'dimension': {'hostname': 'foo1', 'service': 'bar1'}}}


def issue(component=True, custom_config=False):

    issue = {'fields': {'description': 'Monasca alarm',
                        'issuetype': {'name': 'Bug'},
                        'project': {'key': 'MyProject'},
                        'summary': 'Monasca alarm for alarm_defintion test Alarm status '
                                   'changed to ALARM for the alarm_id 0'}}
    if component:
        issue['fields'].update({'components': [{'name': 'MyComponent'}]})
    if custom_config:
        alarm_value = alarm()
        issue['fields'].update({'description': alarm_value.get('alarmName')})
        summary_format_string = 'Alarm created for {0} with severity {1} for {2}'
        summary = summary_format_string.format(alarm_value.get('alarmName'),
                                               alarm_value.get('newState'),
                                               alarm_value.get('alarmId'))
        issue['fields'].update({'summary': summary})
    return issue


class RequestsResponse(object):
    def __init__(self, status):
        self.status_code = status


class TestJira(base.BaseTestCase):

    default_address = 'http://test.jira:3333/?project=MyProject' \
                      '&component=MyComponent'
    default_transitions_value = [[{'id': 100, 'name': 'reopen'}]]
    issue_status_resolved = 'resolved'

    def setUp(self):
        super(TestJira, self).setUp()

        self._trap = queue.Queue()

        mock_log = mock.Mock()
        mock_log.info = self._trap.put
        mock_log.debug = self._trap.put
        mock_log.warn = self._trap.put
        mock_log.error = self._trap.put
        mock_log.exception = self._trap.put

        self._jr = jira_notifier.JiraNotifier(mock_log)

        self._jira_config = {'user': 'username',
                             'password': 'password'}

    @mock.patch('monasca_notification.plugins.jira_notifier.jira')
    def _notify(self,
                transitions_value,
                issue_status,
                jira_config,
                address,
                mock_jira):
        self._jr.config(jira_config)
        alarm_dict = alarm()

        mock_jira_obj = mock.Mock()
        mock_jira.JIRA.return_value = mock_jira_obj
        mock_jira_issue = mock.Mock()
        if issue_status:
            mock_jira_obj.search_issues.return_value = [mock_jira_issue]
            mock_jira_issue.fields.status.name = issue_status
        else:
            mock_jira_obj.search_issues.return_value = []
        mock_jira_obj.transitions.side_effect = transitions_value

        notification = m_notification.Notification(0, 'jira',
                                                   'jira notification',
                                                   address, 0, 0, alarm_dict)

        return mock_jira, mock_jira_obj, self._jr.send_notification(notification)

    def test_send_notification_issue_status_resolved(self):
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        TestJira.issue_status_resolved,
                                                        self._jira_config,
                                                        TestJira.default_address)
        self.assertTrue(result)
        mock_jira_obj.create_issue.assert_not_called()
        self.assertEqual(
            "project=MyProject and reporter='username' and summary ~ '0'",
            mock_jira_obj.search_issues.call_args[0][0])
        self.assertEqual(100, mock_jira_obj.transition_issue.call_args[0][1])

    def test_send_notification_issue_status_closed(self):
        issue_status = 'closed'
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        issue_status,
                                                        self._jira_config,
                                                        TestJira.default_address)
        self.assertTrue(result)
        mock_jira_obj.create_issue.assert_not_called()
        self.assertEqual(
            "project=MyProject and reporter='username' and summary ~ '0'",
            mock_jira_obj.search_issues.call_args[0][0])
        self.assertEqual(100, mock_jira_obj.transition_issue.call_args[0][1])

    def test_send_notification_issue_status_progress(self):
        issue_status = 'progress'
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        issue_status,
                                                        self._jira_config,
                                                        TestJira.default_address)
        self.assertTrue(result)
        mock_jira_obj.create_issue.assert_not_called()
        mock_jira_obj.transitions.assert_not_called()

    def test_send_notification_not_allow_transitions(self):
        transitions_value = [[{'id': 100, 'name': 'not open'}]]
        mock_jira, mock_jira_obj, result = self._notify(transitions_value,
                                                        TestJira.issue_status_resolved,
                                                        self._jira_config,
                                                        TestJira.default_address)
        self.assertTrue(result)
        mock_jira_obj.create_issue.assert_not_called()
        mock_jira_obj.transition_issue.assert_not_called()

    def test_send_notification_without_issue_list(self):
        issue_status = None
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        issue_status,
                                                        self._jira_config,
                                                        TestJira.default_address)
        self.assertTrue(result)
        mock_jira_obj.transitions.assert_not_called()
        self.assertEqual(issue(), mock_jira_obj.create_issue.call_args[1])

    def test_send_notification_without_componet(self):
        issue_status = None
        address = 'http://test.jira:3333/?project=MyProject'
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        issue_status,
                                                        self._jira_config,
                                                        address)
        self.assertTrue(result)
        self.assertEqual(issue(component=False), mock_jira_obj.create_issue.call_args[1])
        mock_jira_obj.transitions.assert_not_called()

    def test_send_notification_create_jira_object_args(self):
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        TestJira.issue_status_resolved,
                                                        self._jira_config,
                                                        TestJira.default_address)
        self.assertEqual('http://test.jira:3333/', mock_jira.JIRA.call_args[0][0])
        self.assertEqual(('username', 'password'), mock_jira.JIRA.call_args[1].get('basic_auth'))
        self.assertEqual(None, mock_jira.JIRA.call_args[1].get('proxies'))

    def test_send_notification_with_proxy(self):
        jira_config = self._jira_config
        jira_config.update({'proxy': 'http://yourid:password@proxyserver:8080'})
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        TestJira.issue_status_resolved,
                                                        jira_config,
                                                        TestJira.default_address)
        self.assertTrue(result)
        self.assertEqual({'https': 'http://yourid:password@proxyserver:8080'},
                         mock_jira.JIRA.call_args[1].get('proxies'))

    def test_send_notification_custom_config_success(self):
        issue_status = None
        jira_config = self._jira_config
        jira_config.update(
            {'custom_formatter': 'tests/resources/test_jiraformat.yml'})
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        issue_status,
                                                        jira_config,
                                                        TestJira.default_address)
        self.assertTrue(result)
        mock_jira_obj.transitions.assert_not_called()
        self.assertEqual(issue(custom_config=True), mock_jira_obj.create_issue.call_args[1])

    def test_send_notification_custom_config_failed(self):
        jira_config = self._jira_config
        jira_config.update(
            {'custom_formatter': 'tests/resources/test_jiraformat_without_summary.yml'})
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        TestJira.issue_status_resolved,
                                                        jira_config,
                                                        TestJira.default_address)
        self.assertFalse(result)

    def test_send_notification_custom_config_without_comments(self):
        jira_config = self._jira_config
        jira_config.update(
            {'custom_formatter': 'tests/resources/test_jiraformat_without_comments.yml'})
        mock_jira, mock_jira_obj, result = self._notify(TestJira.default_transitions_value,
                                                        TestJira.issue_status_resolved,
                                                        jira_config,
                                                        TestJira.default_address)
        self.assertTrue(result)
        mock_jira_obj.add_comment.assert_not_called()

    def test_send_notification_custom_config_exception(self):
        jira_config = self._jira_config
        jira_config.update({'custom_formatter': 'tests/resources/not_exist_file.yml'})
        self.assertRaises(Exception, self._notify,
                          TestJira.default_transitions_value,
                          TestJira.issue_status_resolved,
                          jira_config,
                          TestJira.default_address)

    def test_type(self):
        self.assertEqual('jira', self._jr.type)

    def test_statsd_name(self):
        self.assertEqual('jira_notifier', self._jr.statsd_name)

    def test_config_exception(self):
        self.assertRaises(Exception, self._jr.config, {})



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jinja2 import Template
import jira
from six.moves import urllib
import ujson as json
import yaml

from monasca_notification.plugins.abstract_notifier import AbstractNotifier

"""
   Note:
     This plugin doesn't support multi tenancy. Multi tenancy requires support for
     multiple JIRA server url. JIRA doesn't support OAUTH2 tokens, we may need to get
     the user credentials in query params and store them in monasca DB which we don't want to do.
     That is the reason for not supporting true multitenancy.

    MultiTenancy can be achieved by creating issues in different project for different tenant on
    the same JIRA server.

   notification.address = https://<jira_url>/?project=<project_name>

   Dependency for Jira
       1) Jira plugin requires Jira library. Consumers need to install
         JIRA via pip
       2) (i.e) pip install jira

   Jira Configuration
       1) jira:
            user: username
            password: password

    Sample notification:
       monasca notification-create MyIssuer JIRA https://jira.hpcloud.net/?project=MyProject
       monasca notification-create MyIssuer1 JIRA https://jira.hpcloud.net/?project=MyProject&
                           component=MyComponent

"""


class JiraNotifier(AbstractNotifier):

    _search_query = search_query = "project={} and reporter='{}' and summary ~ '{}'"

    def __init__(self, log):
        self._log = log
        self.jira_fields_format = None

    def config(self, config_dict):
        self._config = {'timeout': 5}
        if not config_dict.get("user") and not config_dict.get("password"):
            message = "Missing user and password settings in JIRA plugin configuration"
            self._log.exception(message)
            raise Exception(message)

        self._config.update(config_dict)
        self.jira_fields_format = self._get_jira_custom_format_fields()

    @property
    def type(self):
        return "jira"

    @property
    def statsd_name(self):
        return 'jira_notifier'

    def _get_jira_custom_format_fields(self):
        jira_fields_format = None

        if (not self.jira_fields_format and self._config.get("custom_formatter")):
            try:
                with open(self._config.get("custom_formatter")) as f:
                    jira_fields_format = yaml.safe_load(f)
            except Exception:
                self._log.exception("Unable to read custom_formatter file. Check file location")
                raise
            # Remove the top element
            jira_fields_format = jira_fields_format["jira_format"]

        return jira_fields_format

    def _build_custom_jira_message(self, notification, jira_fields_format):

        jira_fields = {}

        # Templatize the message object
        jira_field_summary_field = jira_fields_format.get("summary", None)
        if jira_field_summary_field:
            template = Template(jira_field_summary_field)
            jira_fields["summary"] = template.render(notification=notification)

        jira_field_comments_field = jira_fields_format.get("comments", None)
        if jira_field_comments_field:
            template = Template(jira_field_comments_field)
            jira_fields["comments"] = template.render(notification=notification)

        jira_field_description_field = jira_fields_format.get("description", None)
        if jira_field_description_field:
            template = Template(jira_field_description_field)
            jira_fields["description"] = template.render(notification=notification)

        return jira_fields

    def _build_default_jira_message(self, notification):
        """Builds jira message body
        """
        body = {'alarm_id': notification.alarm_id,
                'alarm_definition_id': notification.raw_alarm['alarmDefinitionId'],
                'alarm_name': notification.alarm_name,
                'alarm_description': notification.raw_alarm['alarmDescription'],
                'alarm_timestamp': notification.alarm_timestamp,
                'state': notification.state,
                'old_state': notification.raw_alarm['oldState'],
                'message': notification.message,
                'tenant_id': notification.tenant_id,
                'metrics': notification.metrics}

        jira_fields = {}
        summary_format_string = "Monasca alarm for alarm_defintion {0} status changed to {1} for the alarm_id {2}"
        jira_fields["summary"] = summary_format_string.format(notification.alarm_name,
                                                              notification.state,
                                                              notification.alarm_id)
        jira_fields["comments"] = "{code}%s{code}" % (json.dumps(body, indent=3))
        jira_fields["description"] = 'Monasca alarm'

        return jira_fields

    def _build_jira_message(self, notification):
        if self._config.get("custom_formatter"):
            return self._build_custom_jira_message(notification, self.jira_fields_format)

        return self._build_default_jira_message(notification)

    def send_notification(self, notification):
        """Creates or Updates an issue in Jira
        """

        jira_fields = self._build_jira_message(notification)

        parsed_url = urllib.parse.urlsplit(notification.address)
        query_params = urllib.parse.parse_qs(parsed_url.query)
        # URL without query params
        url = urllib.parse.urljoin(notification.address, urllib.parse.urlparse(notification.address).path)

        jira_fields["project"] = query_params["project"][0]
        if query_params.get("component"):
            jira_fields["component"] = query_params["component"][0]

        auth = (self._config["user"], self._config["password"])
        proxyDict = None
        if (self._config.get("proxy")):
            proxyDict = {"https": self._config.get("proxy")}

        try:
            jira_obj = jira.JIRA(url, basic_auth=auth, proxies=proxyDict)

            self.jira_workflow(jira_fields, jira_obj, notification)
        except Exception:
            self._log.exception("Error creating issue in Jira at URL {}".format(url))
            return False

        return True

    def jira_workflow(self, jira_fields, jira_obj, notification):
        """How does Jira plugin work?
           1) Check whether the issue with same description exists?
           2) If issue exists, and if it is closed state, open it
           3) if the issue doesn't exist, then create the issue
           4) Add current alarm details in comments
        """

        issue_dict = {'project': {'key': jira_fields["project"]},
                      'summary': jira_fields["summary"],
                      'description': jira_fields["description"],
                      'issuetype': {'name': 'Bug'}, }

        # If the JIRA workflow is created with mandatory components
        if jira_fields.get("component"):
            issue_dict["components"] = [{"name": jira_fields.get("component")}]

        search_term = self._search_query.format(issue_dict["project"]["key"],
                                                self._config["user"], notification.alarm_id)
        issue_list = jira_obj.search_issues(search_term)
        if not issue_list:
            self._log.debug("Creating an issue with the data {}".format(issue_dict))
            issue = jira_obj.create_issue(fields=issue_dict)
        else:
            issue = issue_list[0]
            self._log.debug("Found an existing issue {} for this notification".format(issue))
            current_state = issue.fields.status.name
            if current_state.lower() in ["resolved", "closed"]:
                # Open the the issue
                transitions = jira_obj.transitions(issue)
                allowed_transistions = [(t['id'], t['name']) for t in transitions if "reopen" in t['name'].lower()]
                if allowed_transistions:
                    # Reopen the issue
                    jira_obj.transition_issue(issue, allowed_transistions[0][0])

        jira_comment_message = jira_fields.get("comments")
        if jira_comment_message:
            jira_obj.add_comment(issue, jira_comment_message)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

import mock

from monasca_notification.common.repositories.orm import orm_repo
from oslotest import base


class TestOrmRepo(base.BaseTestCase):
    @mock.patch('monasca_notification.common.repositories.orm.orm_repo.engine_from_config')
    def setUp(self, mock_sql_engine_from_config):
        super(TestOrmRepo, self).setUp()
        config = {'database':
                  {'orm':
                   {'url': 'mysql+pymysql://user:password@hostname:3306/mon'}}}
        self._rep = orm_repo.OrmRepo(config)
        self.mock_conn = \
            self._rep._orm_engine.connect.return_value.__enter__.return_value

    def test_fetch_notifications_success(self):
        alarmdef_id = 'alarmdef-123'
        new_state = 'alarm'

        alarm = {'alarmDefinitionId': alarmdef_id,
                 'newState': new_state}
        self.mock_conn.execute.return_value = [('notification-123',
                                                'EMAIL',
                                                'notification-name',
                                                'testaddress',
                                                0)]

        self.assertEqual([('notification-123',
                           'email',
                           'notification-name',
                           'testaddress',
                           0)],
                         self._rep.fetch_notifications(alarm))

        self.mock_conn.execute.assert_called_once()
        self.assertEqual(self._rep._orm_query,
                         self.mock_conn.execute.call_args_list[0][0][0])
        self.assertEqual({'alarm_definition_id': alarmdef_id,
                          'alarm_state': new_state},
                         self.mock_conn.execute.call_args_list[0][1])

    def test_get_alarm_current_state_success(self):
        alarm_id = 'alarm-123'
        alarm_state = 'alarm'
        self.mock_conn.execute.return_value.fetchone.return_value = [alarm_state]

        self.assertEqual(self._rep.get_alarm_current_state(alarm_id), alarm_state)

        self.mock_conn.execute.assert_called_once()
        self.assertEqual(self._rep._orm_get_alarm_state,
                         self.mock_conn.execute.call_args_list[0][0][0])
        self.assertEqual({'alarm_id': alarm_id},
                         self.mock_conn.execute.call_args_list[0][1])

    def test_fetch_notification_method_types_success(self):
        notification_methods = [('EMAIL',), ('WEBHOOK',)]
        self.mock_conn.execute.return_value.fetchall.return_value = notification_methods

        self.assertEqual(self._rep.fetch_notification_method_types(), ['EMAIL', 'WEBHOOK'])

        self.mock_conn.execute.assert_called_once()
        self.assertEqual(self._rep._orm_nmt_query,
                         self.mock_conn.execute.call_args_list[0][0][0])

    def test_insert_notification_method_types_success(self):
        notification_types = ['SLACK', 'HIPCHAT', 'JIRA']
        self.mock_conn.execute.return_value = 1

        self._rep.insert_notification_method_types(notification_types)

        self.assertEqual(self._rep._orm_add_notification_type,
                         self.mock_conn.execute.call_args_list[0][0][0])
        self.assertEqual({'b_name': 'SLACK'},
                         self.mock_conn.execute.call_args_list[0][1])

    def test_get_notification_success(self):
        notification_id = 'notification-123'
        self.mock_conn.execute.return_value.fetchone.return_value = [
            'notification-123',
            'email',
            'notification-name',
            'testaddress',
            0]

        self.assertEqual(['notification-123',
                          'email',
                          'notification-name',
                          'testaddress'],
                         self._rep.get_notification(notification_id))

        self.mock_conn.execute.assert_called_once()
        self.assertEqual(self._rep._orm_get_notification,
                         self.mock_conn.execute.call_args_list[0][0][0])
        self.assertEqual({'notification_id': notification_id},
                         self.mock_conn.execute.call_args_list[0][1])



# (C) Copyright 2015 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import unittest

import pymysql

from monasca_notification.common.repositories import exceptions as exc
from monasca_notification.common.repositories.mysql import mysql_repo


class TestMySqlRepo(unittest.TestCase):
    @mock.patch('monasca_notification.common.repositories.mysql.mysql_repo.pymysql')
    def testReconnect(self, mock_mysql):
        m = mock.MagicMock()

        m.cursor.side_effect = pymysql.Error

        mock_mysql.connect.return_value = m
        mock_mysql.Error = pymysql.Error

        config = {'mysql': {'host': 'foo',
                            'port': '3306',
                            'user': 'bar',
                            'passwd': '1',
                            'db': '2'}}

        repo = mysql_repo.MysqlRepo(config)

        alarm = {'alarmDefinitionId': 'foo',
                 'newState': 'bar'}

        def get_notification(repo, alarm):
            g = repo.fetch_notifications(alarm)
            for x in g:
                pass

        try:
            get_notification(repo, alarm)
        except exc.DatabaseException:
            try:
                get_notification(repo, alarm)
            except Exception:
                pass

        self.assertEqual(mock_mysql.connect.call_count, 2)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import mock
import time
import unittest

from monasca_notification import notification as m_notification
from monasca_notification.types import notifiers


def alarm(metrics):
    return {"tenantId": "0",
            "alarmId": "0",
            "alarmName": "test Alarm",
            "oldState": "OK",
            "newState": "ALARM",
            "severity": "LOW",
            "link": "some-link",
            "lifecycleState": "OPEN",
            "stateChangeReason": "I am alarming!",
            "timestamp": time.time(),
            "metrics": metrics}


class NotifyStub(object):
    def __init__(self, trap, config, send, failure):
        self.config_exception = config
        self.send_exception = send
        self.failure = failure
        self.trap = trap

    @property
    def type(self):
        return "email"

    @property
    def statsd_name(self):
        return "smtp_sent"

    def config(self, config_dict):
        if self.config_exception:
            raise Exception
        else:
            pass

    def send_notification(self, notification_obj):
        if self.send_exception:
            raise Exception
        else:
            if self.failure:
                return False
            else:
                return True


class Statsd(object):
    def __init__(self):
        self.timer = StatsdTimer()
        self.counter = StatsdCounter()

    def get_timer(self):
        return self.timer

    def get_counter(self, key):
        return self.counter


class StatsdTimer(object):
    def __init__(self):
        self.timer_calls = {}

    @contextlib.contextmanager
    def time(self, key):
        self.start(key)
        yield
        self.stop(key)

    def start(self, key):
        key = key + "_start"
        if key in self.timer_calls:
            self.timer_calls[key] += 1
        else:
            self.timer_calls[key] = 1

    def stop(self, key):
        key = key + "_stop"
        if key in self.timer_calls:
            self.timer_calls[key] += 1
        else:
            self.timer_calls[key] = 1


class StatsdCounter(object):
    def __init__(self):
        self.counter = 0

    def increment(self, val):
        self.counter += val


class TestInterface(unittest.TestCase):
    def setUp(self):
        self.trap = []
        self.statsd = Statsd()
        self.email_config = {'server': 'my.smtp.server',
                             'port': 25,
                             'user': None,
                             'password': None,
                             'timeout': 60,
                             'from_addr': 'hpcs.mon@hp.com'}

    def tearDown(self):
        notifiers.possible_notifiers = []
        notifiers.configured_notifiers = {}
        self.trap = []

    def _configExceptionStub(self, log):
        return NotifyStub(self.trap, True, False, False)

    def _sendExceptionStub(self, log):
        return NotifyStub(self.trap, False, True, False)

    def _sendFailureStub(self, log):
        return NotifyStub(self.trap, False, False, True)

    def _goodSendStub(self, log):
        return NotifyStub(self.trap, False, False, False)

    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_enabled_notifications(self, mock_log, mock_smtp):
        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'},
                       'pagerduty': {'address': 'xyz.com'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)
        notifications = notifiers.enabled_notifications()

        self.assertEqual(len(notifications), 3)
        self.assertEqual(sorted(notifications),
                         ["EMAIL", "PAGERDUTY", "WEBHOOK"])

    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_config_missing_data(self, mock_log, mock_smtp):
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.info = self.trap.append

        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)

        self.assertIn("No config data for type: pagerduty", self.trap)

    @mock.patch('monasca_notification.types.notifiers.email_notifier')
    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_config_exception(self, mock_log, mock_smtp, mock_email):
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.exception = self.trap.append

        mock_email.EmailNotifier = self._configExceptionStub

        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'},
                       'pagerduty': {'address': 'abc'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)

        self.assertIn("config exception for email", self.trap)

    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_config_correct(self, mock_log, mock_smtp):
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.info = self.trap.append

        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'},
                       'pagerduty': {'address': 'abc'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)

        self.assertIn("email notification ready", self.trap)
        self.assertIn("webhook notification ready", self.trap)
        self.assertIn("pagerduty notification ready", self.trap)

    @mock.patch('monasca_notification.types.notifiers.email_notifier')
    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_send_notification_exception(self, mock_log, mock_smtp, mock_email):
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.exception = self.trap.append

        mock_email.EmailNotifier = self._sendExceptionStub

        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'},
                       'pagerduty': {'address': 'abc'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)

        notifications = []
        notifications.append(m_notification.Notification(0, 'email', 'email notification',
                                                         'me@here.com', 0, 0, alarm({})))

        notifiers.send_notifications(notifications)

        self.assertIn("send_notification exception for email", self.trap)

    @mock.patch('monasca_notification.types.notifiers.email_notifier')
    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_send_notification_failure(self, mock_log, mock_smtp, mock_email):
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.exception = self.trap.append

        mock_email.EmailNotifier = self._sendFailureStub

        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'},
                       'pagerduty': {'address': 'abc'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)

        notifications = []
        notifications.append(m_notification.Notification(0, 'email', 'email notification',
                                                         'me@here.com', 0, 0, alarm({})))

        sent, failed, invalid = notifiers.send_notifications(notifications)

        self.assertEqual(sent, [])
        self.assertEqual(len(failed), 1)
        self.assertEqual(invalid, [])

    @mock.patch('monasca_notification.types.notifiers.email_notifier')
    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_send_notification_unconfigured(self, mock_log, mock_smtp, mock_email):
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append
        mock_log.info = self.trap.append

        mock_email.EmailNotifier = self._sendExceptionStub

        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)

        self.assertIn("No config data for type: pagerduty", self.trap)

        notifications = []
        notifications.append(m_notification.Notification(0, 'pagerduty', 'pagerduty notification',
                                                         'me@here.com', 0, 0, alarm({})))

        sent, failed, invalid = notifiers.send_notifications(notifications)

        self.assertEqual(sent, [])
        self.assertEqual(failed, [])
        self.assertEqual(len(invalid), 1)

        self.assertIn("attempting to send unconfigured notification: pagerduty", self.trap)

    @mock.patch('monasca_notification.types.notifiers.time')
    @mock.patch('monasca_notification.types.notifiers.email_notifier')
    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_send_notification_correct(self, mock_log, mock_smtp,
                                       mock_email, mock_time):
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append

        mock_email.EmailNotifier = self._goodSendStub

        mock_time.time.return_value = 42

        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'},
                       'pagerduty': {'address': 'abc'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)

        notifications = []
        notifications.append(m_notification.Notification(0, 'email', 'email notification',
                                                         'me@here.com', 0, 0, alarm({})))
        notifications.append(m_notification.Notification(1, 'email', 'email notification',
                                                         'foo@here.com', 0, 0, alarm({})))
        notifications.append(m_notification.Notification(2, 'email', 'email notification',
                                                         'bar@here.com', 0, 0, alarm({})))

        sent, failed, invalid = notifiers.send_notifications(notifications)

        self.assertEqual(len(sent), 3)
        self.assertEqual(failed, [])
        self.assertEqual(invalid, [])

        for n in sent:
            self.assertEqual(n.notification_timestamp, 42)

    @mock.patch('monasca_notification.types.notifiers.email_notifier')
    @mock.patch('monasca_notification.types.notifiers.email_notifier.smtplib')
    @mock.patch('monasca_notification.types.notifiers.log')
    def test_statsd(self, mock_log, mock_smtp, mock_email):
        mock_log.warn = self.trap.append
        mock_log.error = self.trap.append

        mock_email.EmailNotifier = self._goodSendStub

        config_dict = {'email': self.email_config,
                       'webhook': {'address': 'xyz.com'},
                       'pagerduty': {'address': 'abc'}}

        notifiers.init(self.statsd)
        notifiers.config(config_dict)

        notifications = []
        notifications.append(m_notification.Notification(0, 'email', 'email notification',
                                                         'me@here.com', 0, 0, alarm({})))
        notifications.append(m_notification.Notification(1, 'email', 'email notification',
                                                         'foo@here.com', 0, 0, alarm({})))
        notifications.append(m_notification.Notification(2, 'email', 'email notification',
                                                         'bar@here.com', 0, 0, alarm({})))

        notifiers.send_notifications(notifications)

        self.assertEqual(self.statsd.timer.timer_calls['email_time_start'], 3)
        self.assertEqual(self.statsd.timer.timer_calls['email_time_stop'], 3)
        self.assertEqual(self.statsd.counter.counter, 3)

    def test_plugin_load(self):
        config_dict = {"plugins": ["monasca_notification.plugins.hipchat_notifier:HipChatNotifier",
                                   "monasca_notification.plugins.slack_notifier:SlackNotifier"]}

        notifiers.init(self.statsd)
        notifiers.load_plugins(config_dict)
        self.assertEqual(len(notifiers.possible_notifiers), 5)

        configured_plugins = ["email", "webhook", "pagerduty", "hipchat", "slack"]
        for plugin in notifiers.configured_notifiers:
            self.asssertIn(plugin.type in configured_plugins)

    @mock.patch('monasca_notification.types.notifiers.log')
    def test_invalid_plugin_load_exception_ignored(self, mock_log):
        mock_log.exception = self.trap.append
        config_dict = {"plugins": ["monasca_notification.plugins.hipchat_notifier:UnknownPlugin",
                                   "monasca_notification.plugins.slack_notifier:SlackNotifier"]}

        notifiers.init(self.statsd)
        notifiers.load_plugins(config_dict)
        self.assertEqual(len(notifiers.possible_notifiers), 4)
        self.assertEqual(len(self.trap), 1)

        configured_plugins = ["email", "webhook", "pagerduty", "slack"]
        for plugin in notifiers.configured_notifiers:
            self.asssertIn(plugin.type in configured_plugins)

    @mock.patch('monasca_notification.types.notifiers.log')
    def test_no_plugins_keyword_in_possible_notifiers(self, mock_log):
        mock_log.warning = self.trap.append
        config_dict = {"plugins": ["monasca_notification.plugins.slack_notifier:SlackNotifier"],
                       "fake_notifier": ["monasca_notification.plugins.fake_notifier:FakeNotifier"],
                       "slack": {"timeout": 5, "ca_certs": "/etc/ssl/certs/ca-certificates.crt", "insecure": False}
                       }
        notifiers.init(self.statsd)
        notifiers.possible_notifiers = []
        notifiers.load_plugins(config_dict)
        notifiers.config(config_dict)
        self.assertEqual('No notifiers found for fake_notifier', mock_log.warn.call_args[0][0])



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

from monasca_common.kafka import consumer
from monasca_common.kafka import producer
from monasca_notification.common.utils import get_statsd_client
from monasca_notification.processors.alarm_processor import AlarmProcessor
from monasca_notification.processors.notification_processor import NotificationProcessor

log = logging.getLogger(__name__)


class NotificationEngine(object):
    def __init__(self, config):
        self._topics = {}
        self._topics['notification_topic'] = config['kafka']['notification_topic']
        self._topics['retry_topic'] = config['kafka']['notification_retry_topic']

        self._statsd = get_statsd_client(config)
        self._consumer = consumer.KafkaConsumer(
            config['kafka']['url'],
            config['zookeeper']['url'],
            config['zookeeper']['notification_path'],
            config['kafka']['group'],
            config['kafka']['alarm_topic'])
        self._producer = producer.KafkaProducer(config['kafka']['url'])
        self._alarm_ttl = config['processors']['alarm']['ttl']
        self._alarms = AlarmProcessor(self._alarm_ttl, config)
        self._notifier = NotificationProcessor(config)

        self._config = config

    def _add_periodic_notifications(self, notifications):
        for notification in notifications:
            topic = notification.periodic_topic
            if topic in self._config['kafka']['periodic'] and notification.type == "webhook":
                notification.notification_timestamp = time.time()
                self._producer.publish(self._config['kafka']['periodic'][topic],
                                       [notification.to_json()])

    def run(self):
        finished_count = self._statsd.get_counter(name='alarms_finished_count')
        for alarm in self._consumer:
            log.debug('Received alarm >|%s|<', str(alarm))
            notifications, partition, offset = self._alarms.to_notification(alarm)
            if notifications:
                self._add_periodic_notifications(notifications)

                sent, failed = self._notifier.send(notifications)
                self._producer.publish(self._topics['notification_topic'],
                                       [i.to_json() for i in sent])
                self._producer.publish(self._topics['retry_topic'],
                                       [i.to_json() for i in failed])

            self._consumer.commit()
            finished_count.increment()



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
# Copyright 2016 FUJITSU LIMITED

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import monascastatsd

from monasca_common.simport import simport

from monasca_notification.common.repositories import exceptions
from monasca_notification.notification import Notification

log = logging.getLogger(__name__)

NOTIFICATION_DIMENSIONS = {'service': 'monitoring',
                           'component': 'monasca-notification'}


def get_db_repo(config):
    if 'database' in config and 'repo_driver' in config['database']:
        return simport.load(config['database']['repo_driver'])(config)
    else:
        return simport.load('monasca_notification.common.repositories.mysql.mysql_repo:MysqlRepo')(config)


def construct_notification_object(db_repo, notification_json):
    try:
        notification = Notification(notification_json['id'],
                                    notification_json['type'],
                                    notification_json['name'],
                                    notification_json['address'],
                                    notification_json['period'],
                                    notification_json['retry_count'],
                                    notification_json['raw_alarm'])
        # Grab notification method from database to see if it was changed
        stored_notification = grab_stored_notification_method(db_repo, notification.id)
        # Notification method was deleted
        if stored_notification is None:
            log.debug("Notification method {0} was deleted from database. "
                      "Will stop sending.".format(notification.id))
            return None
        # Update notification method with most up to date values
        else:
            notification.name = stored_notification[0]
            notification.type = stored_notification[1]
            notification.address = stored_notification[2]
            notification.period = stored_notification[3]
            return notification
    except exceptions.DatabaseException:
        log.warn("Error querying mysql for notification method. "
                 "Using currently cached method.")
        return notification
    except Exception as e:
        log.warn("Error when attempting to construct notification {0}".format(e))
        return None


def grab_stored_notification_method(db_repo, notification_id):
        try:
            stored_notification = db_repo.get_notification(notification_id)
        except exceptions.DatabaseException:
            log.debug('Database Error.  Attempting reconnect')
            stored_notification = db_repo.get_notification(notification_id)

        return stored_notification


def get_statsd_client(config, dimensions=None):
    local_dims = dimensions.copy() if dimensions else {}
    local_dims.update(NOTIFICATION_DIMENSIONS)
    if 'statsd' in config:
        client = monascastatsd.Client(name='monasca',
                                      host=config['statsd'].get('host', 'localhost'),
                                      port=config['statsd'].get('port', 8125),
                                      dimensions=local_dims)
    else:
        client = monascastatsd.Client(name='monasca',
                                      dimensions=local_dims)
    return client
